// @flow

import React from 'react';
import { wait, render, fireEvent } from 'react-testing-library';
import { loadPlugins, Slot } from 'react-plugin';
import { updateState } from 'react-cosmos-shared2/util';
import {
  cleanup,
  getPluginState,
  mockConfig,
  mockState,
  mockMethod
} from '../../../testHelpers/plugin';
import { DEFAULT_VIEWPORT, getResponsiveViewportStorageKey } from '../shared';
import { register } from '..';

afterEach(cleanup);

const storageKey = getResponsiveViewportStorageKey('mockProjectId');

function registerTestPlugins() {
  register();
  mockConfig('core', { projectId: 'mockProjectId' });
  mockState('rendererCoordinator', { fixtureState: null });
  mockMethod('rendererCoordinator.isValidFixtureSelected', () => true);
}

function loadTestPlugins() {
  loadPlugins();

  return render(<Slot name="fixtureActions" />);
}

it('sets enabled state', async () => {
  registerTestPlugins();
  mockMethod('storage.getItem', () => {});
  mockMethod('rendererCoordinator.setFixtureState', () => {});

  const { getByText } = loadTestPlugins();
  fireEvent.click(getByText(/responsive/i));

  await wait(() =>
    expect(getPluginState('responsivePreview')).toEqual({
      enabled: true,
      viewport: DEFAULT_VIEWPORT
    })
  );
});

it('sets enabled state with stored viewport', async () => {
  registerTestPlugins();
  mockMethod('rendererCoordinator.setFixtureState', () => {});

  const storage = {
    [storageKey]: { width: 420, height: 420 }
  };
  mockMethod('storage.getItem', (context, key) =>
    Promise.resolve(storage[key])
  );

  const { getByText } = loadTestPlugins();
  fireEvent.click(getByText(/responsive/i));

  await wait(() =>
    expect(getPluginState('responsivePreview')).toEqual({
      enabled: true,
      viewport: { width: 420, height: 420 }
    })
  );
});

it('sets viewport in fixture state', async () => {
  registerTestPlugins();
  mockMethod('storage.getItem', () => {});

  let fixtureState = {};
  const handleSetFixtureState = (context, stateChange) => {
    fixtureState = updateState(fixtureState, stateChange);
  };
  mockMethod('rendererCoordinator.setFixtureState', handleSetFixtureState);

  const { getByText } = loadTestPlugins();
  fireEvent.click(getByText(/responsive/i));

  await wait(() => expect(fixtureState.viewport).toEqual(DEFAULT_VIEWPORT));
});

it('sets disabled state', async () => {
  registerTestPlugins();
  mockMethod('storage.getItem', () => {});
  mockMethod('rendererCoordinator.setFixtureState', () => {});

  const { getByText } = loadTestPlugins();

  const getButton = getByText(/responsive/i);
  fireEvent.click(getButton);
  fireEvent.click(getButton);

  await wait(() =>
    expect(getPluginState('responsivePreview')).toEqual({
      enabled: false,
      viewport: DEFAULT_VIEWPORT
    })
  );
});

it('sets disabled state with stored viewport', async () => {
  registerTestPlugins();
  mockMethod('rendererCoordinator.setFixtureState', () => {});

  const storage = {
    [storageKey]: { width: 420, height: 420 }
  };
  mockMethod('storage.getItem', (context, key) =>
    Promise.resolve(storage[key])
  );

  const { getByText } = loadTestPlugins();
  const getButton = getByText(/responsive/i);
  fireEvent.click(getButton);
  fireEvent.click(getButton);

  await wait(() =>
    expect(getPluginState('responsivePreview')).toEqual({
      enabled: false,
      viewport: { width: 420, height: 420 }
    })
  );
});

it('clears viewport in fixture state', async () => {
  registerTestPlugins();
  mockMethod('storage.getItem', () => {});

  let fixtureState = {};
  const handleSetFixtureState = (context, stateChange) => {
    fixtureState = updateState(fixtureState, stateChange);
  };
  mockMethod('rendererCoordinator.setFixtureState', handleSetFixtureState);

  const { getByText } = loadTestPlugins();
  const getButton = getByText(/responsive/i);
  fireEvent.click(getButton);
  fireEvent.click(getButton);

  await wait(() => expect(fixtureState.viewport).toBe(null));
});
